use clap::Parser;
use egui::IconData;
use libpt::cli::args::{VerbosityLevel, HELP_TEMPLATE};
use libpt::log::trace;

mod roll;
use roll::Roller;

pub const TITLE: &str = "Rollator";

/// Placeholder comment that will show in the help in the CLI
///
/// amogus
#[derive(serde::Deserialize, serde::Serialize, Debug, Clone, Parser)]
#[command(help_template = HELP_TEMPLATE, author, version)]
#[serde(default)] // if we add new fields, give them default values when deserializing old state
pub struct RollatorApp {
    #[clap(skip)]
    label: String,

    #[serde(skip)]
    #[clap(skip)]
    value: f32,

    #[clap(skip)]
    roller: Roller,

    #[serde(skip)]
    #[command(flatten)]
    pub(crate) verbosity: VerbosityLevel,

    /// open the informations window on start
    #[serde(skip)]
    #[arg(short = 'i', long = "info")]
    show_info_window: bool,
}

impl Default for RollatorApp {
    fn default() -> Self {
        Self {
            label: TITLE.to_owned(),
            value: 2.7,
            verbosity: VerbosityLevel::INFO,
            show_info_window: false,
            roller: Roller::default(),
        }
    }
}

impl RollatorApp {
    pub fn restore_state(&mut self, cc: &eframe::CreationContext<'_>) {
        // Load previous app state (if any).
        // Note that you must enable the `persistence` feature for this to work.
        if let Some(storage) = cc.storage {
            let old: Self = eframe::get_value(storage, eframe::APP_KEY).unwrap_or_default();

            self.label = old.label;
            self.value = old.value;
            self.roller = old.roller;
        }
    }

    /// Called once before the first frame.
    #[allow(unused)] // can be used for starting with wasm
    pub fn new() -> Self {
        Default::default()
    }

    pub fn new_with_cli() -> Self {
        Self::parse()
    }

    pub fn info_diag(&mut self, ctx: &egui::Context) {
        trace!("rendering info dialogue");
        ctx.show_viewport_immediate(
            egui::ViewportId::from_hash_of(format!("{TITLE}: Information")),
            egui::ViewportBuilder::default()
                .with_title(format!("{TITLE}: Information"))
                .with_inner_size([500.0, 200.0]),
            |ctx, class| {
                assert!(
                    class == egui::ViewportClass::Immediate,
                    "This egui backend doesn't support multiple viewports"
                );

                egui::CentralPanel::default().show(ctx, |ui| {
                    ui.label(format!("{TITLE} v{}", env!("CARGO_PKG_VERSION")));
                    ui.hyperlink_to("Source Code\n", env!("CARGO_PKG_REPOSITORY"));
                    ui.label(format!("Author: {}", env!("CARGO_PKG_AUTHORS")));
                    ui.label(format!("License: {}", env!("CARGO_PKG_LICENSE")));
                    ui.label(format!(
                        "\n{TITLE} is free software. If you paid for this you were scammed.\n"
                    ));

                    ////////////////////////////////////////////////////////////////

                    ui.horizontal(|ui| {
                        ui.spacing_mut().item_spacing.x = 0.0;
                        ui.label("Powered by ");
                        ui.hyperlink_to("egui", "https://github.com/emilk/egui");
                        ui.label(" and ");
                        ui.hyperlink_to(
                            "eframe",
                            "https://github.com/emilk/egui/tree/master/crates/eframe",
                        );
                        ui.label(".");
                    });

                    ////////////////////////////////////////////////////////////////
                });
                if ctx.input(|i| i.viewport().close_requested()) {
                    // Tell parent viewport that we should not show next frame:
                    self.show_info_window = false;
                }
            },
        );
    }

    fn top_panel(&mut self, ui: &mut egui::Ui, ctx: &egui::Context) {
        egui::menu::bar(ui, |ui| {
            // NOTE: no File->Quit on web pages!
            let is_web = cfg!(target_arch = "wasm32");
            if !is_web {
                ui.menu_button("File", |ui| {
                    if ui.button("Info").clicked() {
                        self.show_info_window = true;
                    }
                    ui.separator();
                    if ui.button("Quit").clicked() {
                        ctx.send_viewport_cmd(egui::ViewportCommand::Close);
                    }
                });
                ui.add_space(16.0);
            }
            if self.show_info_window {
                self.info_diag(ctx);
            }

            egui::widgets::global_dark_light_mode_buttons(ui);
        });
    }

    fn roller_panel(&mut self, ui: &mut egui::Ui, ctx: &egui::Context) {
        ui.heading("Roll");

        ui.vertical_centered_justified(|ui| {
            self.roller.labels(ui);
            self.roller.rolling_space(ui);
        });
    }

    fn meta_panel(&mut self, ui: &mut egui::Ui, ctx: &egui::Context) {
        // The central panel the region left after adding TopPanel's and SidePanel's
        ui.heading(TITLE);

        ui.horizontal(|ui| {
            ui.label("Write something: ");
            ui.text_edit_singleline(&mut self.label);
        });

        ui.add(egui::Slider::new(&mut self.value, 0.0..=10.0).text("value"));
        if ui.button("Increment").clicked() {
            self.value += 1.0;
        }
    }
}

impl eframe::App for RollatorApp {
    /// Called by the frame work to save state before shutdown.
    fn save(&mut self, storage: &mut dyn eframe::Storage) {
        eframe::set_value(storage, eframe::APP_KEY, self);
    }

    /// Called each time the UI needs repainting, which may be many times per second.
    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        egui::TopBottomPanel::top("top_panel").show(ctx, |ui| {
            self.top_panel(ui, ctx);
        });
        egui::TopBottomPanel::top("top_panel2").show(ctx, |ui| {
            self.meta_panel(ui, ctx);
        });
        egui::CentralPanel::default().show(ctx, |ui| {
            ui.heading("idk what to put here lol");
        });
        egui::SidePanel::right(egui::Id::new("roll_panel")).show(ctx, |ui| {
            self.roller_panel(ui, ctx);
        });
        egui::TopBottomPanel::bottom("bot_panel").show(ctx, bottom_label);
    }
}

fn bottom_label(ui: &mut egui::Ui) {
    ui.with_layout(egui::Layout::bottom_up(egui::Align::LEFT), |ui| {
        ui.horizontal(|ui| {
            ui.spacing_mut().item_spacing.x = 0.0;
            ui.label(format!("{TITLE} v{}", env!("CARGO_PKG_VERSION")));
            ui.label(" | Powered by ");
            ui.hyperlink_to("egui", "https://github.com/emilk/egui");
            ui.label(" and ");
            ui.hyperlink_to(
                "eframe",
                "https://github.com/emilk/egui/tree/master/crates/eframe",
            );
            ui.label(".");
        });
        egui::warn_if_debug_build(ui);
    });
}

#[allow(unused)] // it is used in the main function
pub(crate) fn load_icon() -> IconData {
    const ICON_RAW: &[u8; 27618] = include_bytes!("../assets/icon-512.jpg");
    let (icon_rgba, icon_width, icon_height) = {
        let image = image::load_from_memory(ICON_RAW)
            .expect("Failed to open icon path")
            .into_rgba8();
        let (width, height) = image.dimensions();
        let rgba = image.into_raw();
        (rgba, width, height)
    };

    IconData {
        rgba: icon_rgba,
        width: icon_width,
        height: icon_height,
    }
}
