use std::fmt::Display;

use egui::{epaint::PathStroke, *};
use egui::{vec2, Pos2, Rect};
use libpt::log::debug;
use rand::Rng;
use strum::{EnumIter, IntoEnumIterator};

#[derive(
    serde::Deserialize,
    serde::Serialize,
    Debug,
    Clone,
    Default,
    EnumIter,
    Hash,
    PartialEq,
    Eq,
    PartialOrd,
    Ord,
    Copy,
)]
pub enum Dice {
    D2,
    D4,
    D6,
    D8,
    D10,
    D12,
    #[default]
    D20,
    D100,
}

impl From<Dice> for u8 {
    fn from(value: Dice) -> Self {
        match value {
            Dice::D2 => 2,
            Dice::D4 => 4,
            Dice::D6 => 6,
            Dice::D8 => 8,
            Dice::D10 => 10,
            Dice::D12 => 12,
            Dice::D20 => 20,
            Dice::D100 => 100,
        }
    }
}

impl Display for Dice {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Dice::D2 => "D2",
                Dice::D4 => "D4",
                Dice::D6 => "D6",
                Dice::D8 => "D8",
                Dice::D10 => "D10",
                Dice::D12 => "D12",
                Dice::D20 => "D20",
                Dice::D100 => "D100",
            }
        )
    }
}

#[derive(serde::Deserialize, serde::Serialize, Debug, Clone, Default)]
pub struct Roller {
    selected_dice: Dice,
    new: bool,
    value: u8,
}

impl Roller {
    pub(crate) fn roll(&mut self) -> u8 {
        if self.new {
            self.value = rand::thread_rng().gen_range(1..=self.selected_dice.into());
            self.new = false;
        }
        self.value
    }

    pub(crate) fn labels(&mut self, ui: &mut egui::Ui) {
        ui.horizontal(|ui| {
            for variant in Dice::iter() {
                let dice_label =
                    ui.selectable_label(variant == self.selected_dice, variant.to_string());
                if dice_label.clicked() {
                    self.selected_dice = variant;
                }
            }
        });
        ui.separator();
    }

    pub(crate) fn rolling_space(&mut self, ui: &mut egui::Ui) {
        let color = if ui.visuals().dark_mode {
            Color32::from_additive_luminance(196)
        } else {
            Color32::from_black_alpha(240)
        };

        let button = ui.button("> ROLL <");
        if button.clicked() {
            debug!("rolling dice");
            self.new = true;
        }
        ui.add_space(10.0);
        ui.heading(self.roll().to_string());
        ui.add_space(10.0);
        ui.centered_and_justified(|ui| {
            let _canvas = egui::containers::Frame::canvas(ui.style()).show(ui, |ui| {
                ui.ctx().request_repaint();
                let time = ui.input(|i| i.time);

                let desired_size = ui.available_width() * vec2(1.0, 0.35);
                let (_id, rect) = ui.allocate_space(desired_size);

                let to_screen = emath::RectTransform::from_to(
                    Rect::from_x_y_ranges(0.0..=1.0, -1.0..=1.0),
                    rect,
                );
                let mut shapes = vec![];

                for &mode in &[2.0, 3.0, 5.0] {
                    let n = 120;
                    let speed = 1.5;

                    let points: Vec<Pos2> = (0..=n)
                        .map(|i| {
                            let t = i as f64 / (n as f64);
                            let amp = (time * speed * mode).sin() / mode;
                            let y = amp * (t * std::f64::consts::TAU / 2.0 * mode).sin();
                            to_screen * pos2(t as f32, y as f32)
                        })
                        .collect();

                    let thickness = 10.0 / mode as f32;
                    shapes.push(epaint::Shape::line(
                        points,
                        PathStroke::new(thickness, color),
                    ));
                }
                ui.painter().extend(shapes);
            });
        });
    }
}
